{******************************************************************************}
{ Package:      Clinical Case Registries Custom Components                     }
{ Date Created: November 18, 2004                                              }
{ Site Name:    Hines OIFO                                                     }
{ Developers:                                                                  }
{ Description:     ug and troubleshooting classes.                             }
{ Note:                                                                        }
{******************************************************************************}

unit uROR_   ug;

{$I Components.inc}

interface

uses
  Classes, StrUtils, SysUtils, Windows, StdCtrls, Messages;

type

  {=============================== ICCR   ugTrace ==============================
    Overview:     Interface for debug trace objects.
    SeeAlso:      TCCR   ugLog; TCCR   ugTrace; TCCRStr   ugLog 
    Description:
      In order to implement automatic destruction, debug trace objects are
      referenced via the ICCR   ugTrace interface. Use the TCCR   ugTrace.Create
      to create a trace object and get its interface.
      <p>
      If a name is assigned to a trace object (usually a name of the function or
      procedure where the object is created), the object automatically writes
      entry (when created) and exit (when destroyed) messages to a current
      debug log.
      <p>
      You can also store the interface in a local variable and use the Error
      and Write methods to write additional messages to the log between entry
      and exit messages.
      <p>
      See the TCCR   ugTrace class for implemetation details.
  }
  ICCR   ugTrace = interface(IInterface)
    ['{C3A5E15C-2099-4916-801F-4D10339C81EE}']

    // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    procedure Error(const aFormat: String; const Args: array of const); overload;
    {
      Overview:     Writes an error message to a debug log.
      SeeAlso:      ICCR   ugTrace.ErrorPrefix; ICCR   ugTrace.Write
      Keywords:     Error,ICCR   ugTrace
      Description:
        There are 2 overloaded versions of the Error method.

        Use Error(const aFormat: String; const Args: array of const) to format
        a list of paramaters specified by the <i>Args</i> according to the
        <i>aFormat</i> format string and write the resulting error message to
        a debug log. All error messages are prepended with a string stored in
        the ErrorPrefix property.
        
        Use Error(const aMsg: String) to write an error message specified by
        the <i>aMsg</i> parameter to a debug log. All error messages are
        prepended with a string stored in the ErrorPrefix property.
    }
    procedure Error(const aMsg: String); overload;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Returns the value of the EntryPrefix property.
      SeeAlso:      ICCR   ugTrace.EntryPrefix; ICCR   ugTrace.SetEntryPrefix
      Description:
        GetEntryPrefix is the read implementation of the EntryPrefix property.
    }
    function GetEntryPrefix: String;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Returns the value of the ErrorPrefix property.
      SeeAlso:      ICCR   ugTrace.ErrorPrefix; ICCR   ugTrace.SetErrorPrefix
      Description:
        GetErrorPrefix is the read implementation of the ErrorPrefix property.
    }
    function GetErrorPrefix: String;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Returns the value of the ExitPrefix property.
      SeeAlso:      ICCR   ugTrace.ExitPrefix; ICCR   ugTrace.SetExitPrefix
      Description:
        GetExitPrefix is the read implementation of the ExitPrefix property.
    }
    function GetExitPrefix: String;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Returns the value of the Indent property.
      SeeAlso:      ICCR   ugTrace.Indent; ICCR   ugTrace.SetIndent
      Description:
        GetIndent is the read implementation of the Indent property.
    }
    function GetIndent: Integer;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Restores default property values.
      SeeAlso:      ICCR   ugTrace.EntryPrefix; ICCR   ugTrace.ErrorPrefix;
                    ICCR   ugTrace.ExitPrefix; ICCR   ugTrace.Indent;
                    TCCR   ugTrace.RestoreDefaults
      Keywords:     RestoreDefaults,ICCR   ugTrace
      Description:
        Use RestoreDefaults to restore initial values of the trace prefixes
        (EntryPrefix, ErrorPrefix, and ExitPrefix) and indentation level
        (Indent).
        <p>
        See the TCCR   ugTrace class for implemetation details.
    }
    procedure RestoreDefaults;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Sets the value of the EntryPrefix property.
      SeeAlso:      ICCR   ugTrace.EntryPrefix; ICCR   ugTrace.GetEntryPrefix
      Description:
        SetEntryPrefix is the write implementation of the EntryPrefix property.
    }
    procedure SetEntryPrefix(const aValue: String);

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Sets the value of the ErrorPrefix property.
      SeeAlso:      ICCR   ugTrace.ErrorPrefix; ICCR   ugTrace.GetErrorPrefix
      Description:
        SetErrorPrefix is the write implementation of the ErrorPrefix property.
    }
    procedure SetErrorPrefix(const aValue: String);

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Sets the value of the ExitPrefix property.
      SeeAlso:      ICCR   ugTrace.ExitPrefix; ICCR   ugTrace.GetExitPrefix
      Description:
        SetExitPrefix is the write implementation of the ExitPrefix property.
    }
    procedure SetExitPrefix(const aValue: String);

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Sets the value of the Indent property.
      SeeAlso:      ICCR   ugTrace.Indent; ICCR   ugTrace.GetIndent
      Description:
        SetIndent is the write implementation of the Indent property.
    }
    procedure SetIndent(const aValue: Integer);

    // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    procedure Write(const aFormat: String; const Args: array of const); overload;
    {
      Overview:     Writes a message to a debug log.
      SeeAlso:      ICCR   ugTrace.Error
      Keywords:     Write,ICCR   ugTrace
      Description:
        There are 2 overloaded versions of the Write method.

        Use Write(const aMsg: String) to write a message specified by the
        <i>aMsg</i> parameter to a debug log.

        Use Write(const aFormat: String; const Args: array of const) to format
        a list of paramaters specified by the <i>Args</i> according to the
        <i>aFormat</i> format string and write the resulting message to a debug
        log.
    }
    procedure Write(const aMsg: String); overload;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     A prefix that indicates creation of a trace object.
      SeeAlso:      ICCR   ugTrace.ExitPrefix; TCCR   ugTrace.EntryPrefix
      Keywords:     EntryPrefix,ICCR   ugTrace
      Description:
        Use EntryPrefix to get or set the prefix of the entry message that is
        written to a debug log when a named trace object is created.
        <p>
        See the TCCR   ugTrace class for implemetation details.
    }
    property EntryPrefix: String  read GetEntryPrefix  write SetEntryPrefix;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     A prefix that marks an error message.
      SeeAlso:      ICCR   ugTrace.Error; TCCR   ugTrace.ErrorPrefix
      Keywords:     ErrorPrefix,ICCR   ugTrace
      Description:
        Use ErrorPrefix to get or set the prefix for the error messages.
        <p>
        See the TCCR   ugTrace class for implemetation details.
    }
    property ErrorPrefix: String  read GetErrorPrefix  write SetErrorPrefix;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     A prefix that indicates destruction of a trace object.
      SeeAlso:      ICCR   ugTrace.EntryPrefix; TCCR   ugTrace.ExitPrefix
      Keywords:     ExitPrefix,ICCR   ugTrace
      Description:
        Use ExitPrefix to get or set the prefix of the exit message that is
        written to a debug log when a named trace object is destroyed. 
        <p>
        See the TCCR   ugTrace class for implemetation details.
    }
    property ExitPrefix: String  read GetExitPrefix  write SetExitPrefix;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Number of spaces that are added before a message written
                    to a debug log.
      SeeAlso:      ICCR   ugTrace.Error; ICCR   ugTrace.Write;
                    TCCR   ugTrace.Indent
      Keywords:     Indent,ICCR   ugTrace
      Description:
        Indent specifies the number of spaces added before each message written
        to a debug log by the Error and Write methods.
        <p>
        See the TCCR   ugTrace class for implemetation details.
    }
    property Indent: Integer  read GetIndent  write SetIndent;

  end;

  {============================= TCCRCustom   ugLog ============================
    Overview:     Base class for a debug log.
    SeeAlso:      TCCR   ugLog; TCCRStr   ugLog
    Description:
      TCCRCustom   ugLog encapsulates a basic functionality of a debug log.
      Descendants of this classs are thread-safe.
      <p>
      Do not create instances of TCCRCustom   ugLog. To open a log, use a
      descendant of TCCRCustom   ugLog, such as TCCR   ugLog or TCCRStr   ugLog.
      No more than one debug log can exists inside an application.
  }
  TCCRCustom   ugLog = class(TComponent)
  private

    fIndent:   Integer;
    fMaxCount: Integer;
    fMaxSize:  Integer;

    procedure SetIndent(const aValue: Integer);
    procedure SetMaxCount(const aValue: Integer);
    procedure SetMaxSize(const aValue: Integer);

  protected

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Internal method that appends a string to a log.
      SeeAlso:      TCCRCustom   ugLog.Append
      Keywords:     AppendString,TCCRCustom   ugLog
      Description:
        Override AppendString in a descendant class to implement addition of a
        string to a debug log. <i>aText</i> specifies the string.
    }
    procedure AppendString(const aText: String); virtual; abstract;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Truncates a debug log.
      SeeAlso:      TCCRCustom   ugLog.MaxCount; TCCRCustom   ugLog.MaxSize
      Keywords:     Truncate,TCCRCustom   ugLog
      Description:
        This method truncates a debug log according to values of the MaxCount
        and/or MaxSize properties. As implemented in TCCRCustom   ugLog,
        Truncate does nothing.
    }
    procedure Truncate; virtual;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Indentation level for new log lines.
      SeeAlso:      TCCRCustom   ugLog.AppendString
      Keywords:     Indent,TCCRCustom   ugLog
      Description:
        Use Indent to get or set indentation level (in characters) for new log
        lines.
    }
    property Indent: Integer  read fIndent  write SetIndent;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Maximum number of lines in a log.
      SeeAlso:      TCCRCustom   ugLog.MaxSize; TCCRCustom   ugLog.Truncate
      Keywords:     MaxCount,TCCRCustom   ugLog
      Description:
        Use MaxCount to get or set maximum number of lines that can be stored
        in a debug log. When number of lines in a log becomes greater than the
        value of this parameter, the oldest lines are automatically deleted
        from the beginning of the log.
        <p>
        <b>Note:</b> Not all descendant classes support this kind of truncation.
    }
    property MaxCount: Integer  read fMaxCount  write SetMaxCount;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Maximum log size (in bytes).
      SeeAlso:      TCCRCustom   ugLog.MaxCount; TCCRCustom   ugLog.Truncate
      Keywords:     MaxSize,TCCRCustom   ugLog
      Description:
        Use MaxSize to get or set maximum size of a debug log in bytes. When
        the log size becomes greater than the value of this parameter, the
        oldest lines are automatically deleted from the beginning of the log.
        <p>
        <b>Note:</b> Not all descendant classes support this kind of truncation.
    }
    property MaxSize: Integer  read fMaxSize  write SetMaxSize;

  public

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Creates and initializes an instance of TCCRCustom   ugLog.
      SeeAlso:      CCR   ugLog; TComponent.Owner; TObject.Create
      Keywords:     Create,TCCRCustom   ugLog
      Description:
        As implemented in TCCRCustom   ugLog, Create assigns a reference to a
        new instance to an internal variable. If another log object has been
        created before, it is destroyed.
    }
    constructor Create(aDummyParameter: TComponent = nil); override;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Destroys an instance of TCCRCustom   ugLog.
      SeeAlso:      TObject.Free; TComponent.Destroy
      Keywords:     Destroy,TCCRCustom   ugLog
      Description:
        Do not call Destroy directly. Instead, call Free, which checks that the
        TCCRCustom   ugLog reference is not nil before calling Destroy.
    }
    destructor Destroy; override;

    // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    procedure Append(const aStr: String); overload;
    procedure Append(const aFormat: String; const Args: array of const); overload;
    {
      Overview:     Appends content of a string list to a debug log.
      SeeAlso:      TCCRCustom   ugLog.AppendString
      Keywords:     Append,TCCRCustom   ugLog
      Description:
        There are 3 overloaded version of the Append procedure.

        Use Append(const aFormat: String; const Args: array of const) to format
        a list of paramaters specified by the <i>Args</i> according to the
        <i>aFormat</i> format string and add the result to the end of a debug
        log.

        Use Append(const aStr: String) to add a new line specified by the
        <i>aStr</i> parameter to the end of a debug log.

        Use Append(const aText: TStringList) to add items of the <i>aText</i>
        string list to the end of a debug log.
    }
    procedure Append(const aText: TStringList); overload;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Starts a group of updates to a debug log.
      SeeAlso:      TCCRCustom   ugLog.EndUpdate; TCCRCustom   ugLog.Lock
      Keywords:     BeginUpdate,TCCRCustom   ugLog
      Description:
        Call BeginUpdate before adding multiple items to a log. In addition to
        improving the peroformance, this method also protects the log from
        concurrent updates from multiple threads.
        <p>
        Descendant classes that override BeginUpdate should always call the
        inherited method right after entering the procedure.
    }
    procedure BeginUpdate; virtual;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Clears a log.
      Keywords:     Clear,TCCRCustom   ugLog
      Description:
        Use clear to delete all entries from a log. As implemented in
        TCCRCustom   ugLog, Clear does nothing.
        <p>
        <b>Note:</b> Not all descendant classes support this functionality.
    }
    procedure Clear; virtual;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Ends a group of updates to a debug log.
      SeeAlso:      TCCRCustom   ugLog.BeginUpdate; TCCRCustom   ugLog.Unlock
      Keywords:     EndUpdate,TCCRCustom   ugLog
      Description:
        Call EndUpdate after adding multiple items to a log. In addition to
        improving the peroformance, this method also re-enables log updates
        from from other threads.
        <p>
        Descendant classes that override EndUpdate should always call the
        inherited method right before returning from the procedure.
    }
    procedure EndUpdate; virtual;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Prevents other threads from modifying a log.
      SeeAlso:      TCCRCustom   ugLog.Unlock
      Keywords:     Lock,TCCRCustom   ugLog
      Description:
        Place any code that modifies a log object between Lock and Unlock in
        multi-threaded applications.
    }
    class procedure Lock;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Saves a debug log to a file.
      Keywords:     SaveToFile,TCCRCustom   ugLog
      Description:
        Use SaveToFile to write content of a log to a file. <i>aFileName</i>
        specifies a full file name.
        <p>
        <b>Note:</b> Not all descendant classes support this functionality.
    }
    procedure SaveToFile(const aFileName: String); virtual;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Allows unpdating a log in other threads.
      SeeAlso:      TCCRCustom   ugLog.Lock
      Keywords:     Unlock,TCCRCustom   ugLog
      Description:
        Place any code that modifies a log object between Lock and Unlock in
        multi-threaded applications.
    }
    class procedure Unlock;

  end;

  {================================ TCCR   ugLog ===============================
    Overview:        ug log in a debugger window.
    SeeAlso:      TCCRMemo   ugLog; TCCRStr   ugLog
    Description:
      TCCR   ugLog implements a debug log that is shown in a debugger window
      and in the Event Log window of the Delphi IDE. It can also be monitored
      using applications like the freeware    ugView developed by
      <a href="http://www.sysinternals.com">Sysinternals</a>.
  }
  TCCR   ugLog = class(TCCRCustom   ugLog)
  protected

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Internal method that appends a string to a log.
      SeeAlso:      TCCRCustom   ugLog.AppendString
      Keywords:     AppendString,TCCR   ugLog
      Description:
        AppendString uses the Output   ugString procedure (Win32 API) to send a
        string specified by the <i>aText</i> parameter to the debugger for the
        current application.
    }
    procedure AppendString(const aText: String); override;

  end;

  {=============================== TCCR   ugTrace ==============================
    Overview:     Thread-safe implemetation of a trace object.
    SeeAlso:      TCCR   ugLog; TCCR   ugTrace.Create; TCCRStr   ugLog
    Description:
      TCCR   ugTrace is a debug trace object. Usually an instance of this object
      is created in the very beginning of a function/procedure. The object
      writes an entry message to a current debug log when it is created, and an
      exit message when it is automatically destroyed right before exiting from
      the function/procedure.
      <p>
      TCCR   ugTrace implements the ICCR   ugTrace interface. Use the Create
      class method to create a trace object and get its interface.
  }
  TCCR   ugTrace = class(TInterfacedObject,ICCR   ugTrace)
  private

    fEntryPrefix:   String;
    fErrorPrefix:   String;
    fExitPrefix:    String;
    fIndent:        Integer;
    fName:          String;
    fParent:        TCCR   ugTrace;

    function  GetEntryPrefix: String;
    function  GetErrorPrefix: String;
    function  GetExitPrefix: String;
    function  GetIndent: Integer;
    procedure SetEntryPrefix(const aValue: String);
    procedure SetErrorPrefix(const aValue: String);
    procedure SetExitPrefix(const aValue: String);
    procedure SetIndent(const aValue: Integer);

  protected

    // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    procedure Error(const aFormat: String; const Args: array of const); overload;
    {
      Overview:     Implementations of Error methods.
      SeeAlso:      ICCR   ugTrace.Error
      Keywords:     Error,TCCR   ugTrace
      Description:
        See the ICCR   ugTrace interface for details.
    }
    procedure Error(const aMsg: String); overload;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Restores default property values.
      SeeAlso:      TCCR   ugTrace.EntryPrefix; TCCR   ugTrace.ErrorPrefix;
                    TCCR   ugTrace.ExitPrefix; TCCR   ugTrace.Indent;
                    ICCR   ugTrace.RestoreDefaults
      Keywords:     RestoreDefaults,TCCR   ugTrace
      Description:
        RestoreDefaults restores initial values of the trace prefixes
        (EntryPrefix, ErrorPrefix, and ExitPrefix) and indentation level
        (Indent). If a trace object has a parent, corresponding parent's
        values are assigned to object properties. Otherwise, default values
        are assigned as follow:
        <p>
        <p>  EntryPrefix := '>>> ';
        <p>  ErrorPrefix := '! ! !  ';
        <p>  ExitPrefix := '<<< ';
        <p>  Indent := 0;
    }
    procedure RestoreDefaults; virtual;

    // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    procedure Write(const aFormat: String; const Args: array of const); overload;
    {
      Overview:     Implementations of Write methods.
      SeeAlso:      ICCR   ugTrace.Write
      Keywords:     Write,TCCR   ugTrace
      Description:
        See the ICCR   ugTrace interface for details.
    }
    procedure Write(const aMsg: String); overload;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     A prefix that indicates creation of a trace object.
      SeeAlso:      ICCR   ugTrace.EntryPrefix
      Keywords:     EntryPrefix,TCCR   ugTrace
      Description:
        EntryPrefix stores a prefix of the entry message that is written to a
        debug log when a named trace object is created. This message is
        formatted as follow:
        <p>
          <Indent><EntryPrefix><Name>
        <p>
        where <Indent> is amount of space characters specified by the Indent
        property, <EntryPrefix> is the values of the EntryPrefix property, and
        <Name> is the name assigned to the trace object.
        <p>
        The property value is used as default prefix for all subsequent trace
        objects until one of them changes it again. The change does not affect
        trace objects that have been created earlier.
    }
    property EntryPrefix: String  read GetEntryPrefix  write SetEntryPrefix;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     A prefix that marks an error message.
      SeeAlso:      ICCR   ugTrace.ErrorPrefix
      Keywords:     ErrorPrefix,TCCR   ugTrace
      Description:
        Use ErrorPrefix to get or set the prefix for the error messages. The
        value is used as default prefix for all subsequent calls of the Error
        method. The change does not affect trace objects that have been created
        earlier.
    }
    property ErrorPrefix: String  read GetErrorPrefix  write SetErrorPrefix;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     A prefix that indicates destruction of a trace object.
      SeeAlso:      ICCR   ugTrace.ExitPrefix
      Keywords:     ExitPrefix,TCCR   ugTrace
      Description:
        Use ExitPrefix to get or set the prefix of the exit message that is
        written to a debug log when a named trace object is destroyed. This
        message is formatted as follow:
        <p>
          <Indent><ExitPrefix><Name>
        <p>
        where <Indent> is amount of space characters specified by the Indent
        property, <ExitPrefix> is the values of the ExitPrefix property, and
        <Name> is the name assigned to the trace object.
        <p>
        A new property value will be used as default prefix for all subsequent
        trace objects until one of them changes it again.  The change does not
        affect trace objects that have been created earlier.
    }
    property ExitPrefix: String  read GetExitPrefix  write SetExitPrefix;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Number of spaces that are added before a message written
                    to a debug log.
      SeeAlso:      ICCR   ugTrace.Eror; ICCR   ugTrace.Indent;
                    ICCR   ugTrace.Write
      Keywords:     Indent,TCCR   ugTrace
      Description:
        Indent specifies the number of spaces added before each message written
        to a debug log by the Error and Write methods.
        <p>
        When a new trace object is created, it inherits property values from
        its parent (the previous trace object). If a name parameter was
        specified during creation of a trtace object, the entry message is
        written to a log using the parent indentation level. After that, the
        Indent property is automatically incremented by 3. Initial value of the
        Indent property is automatically restored before writing the exit
        message.
        <p>
        If no name is assigned to a trace object, then the object does not
        change the indentation level inherited from its parent.
    }
    property Indent: Integer  read GetIndent  write SetIndent;

  public

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Destroys an instance of TCCR   ugTrace.
      SeeAlso:      TObject.Destroy
      Keywords:     Destroy,TCCR   ugTrace
      Description:
        Do not call Destroy directly.
    }
    destructor Destroy; override;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Creates a trace object and returns its interface.
      Keywords:     Create,TCCR   ugTrace
      Description:
        Use this method to create a trace object and get its interface. A class
        function is used instead of a constructor in order to create an
        interface reference instead of an object reference. This guarantees that
        the object will be automatically destroyed when the reference goes out
        of scope (in the end of the block where the object was created). Use
        <i>aName</i> to specify a name of a trace object (usually a name of the
        function or procedure where the object is created).
        <p>
        If a trace object has a name, it automatically writes an entry message
        when it is created and an exit message when destroyed. It also
        increments the indentation level by 3.
        <p>
        Unnamed trace objects do not display entry/exit messages and do not
        change the indentation level. For example, you can use such object in a
        thread initialization section to modify the default prefixes and/or
        indentation level (to distinguish log entries from different threads).
        <p>
        It is recommended to create no more than one trace object per
        function/procedure (in the very beginning of it).
    }
    class function Create(const aName: String = ''): ICCR   ugTrace;

  end;

  {============================= TCCRDummy   ugLog =============================
    Overview:     Pseudo debug log.
    SeeAlso:      CCR   ugLog
    Description:
      TCCRDummy   ugLog implements a debug log that discards all entries added
      to it. An instance of this class is returned by the CCR   ugLog if no
      other log object has been explicitly created.
  }
  TCCRDummy   ugLog = class(TCCRCustom   ugLog)
  protected

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Internal method that appends a string to a log.
      SeeAlso:      TCCRCustom   ugLog.AppendString
      Keywords:     AppendString,TCCRDummy   ugLog
      Description:
        As implementned in TCCRDummy   ugLog, AppendString does nothing.
    }
    procedure AppendString(const aText: String); override;

  end;

  {============================== TCCRStr   ugLog ==============================
    Overview:     String list implementation of a debug log.
    SeeAlso:      TCCR   ugLog; TCCRMemo   ugLog
    Description:
      TCCRStr   ugLog implements a debug log that is stored in an internal
      string list. Maximum size of the log can be limited by the MaxCount
      property.
  }
  TCCRStr   ugLog = class(TCCRCustom   ugLog)
  private

    fStrLst: TStrings;

  protected

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Internal method that appends a string to a log.
      SeeAlso:      TCCRCustom   ugLog.AppendString
      Keywords:     AppendString,TCCRStr   ugLog
      Description:
        AppendString adds a string specified by the <i>aText</i> parameter to
        the end of the internal string list and calls Truncate to check its
        size.
    }
    procedure AppendString(const aText: String); override;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Specialized AssignTo.
      SeeAlso:      TPersistent.AssignTo
      Keywords:     AssignTo,TCCRStr   ugLog
      Description:
        Do not call protected AssignTo method. The Assign method of a persistent
        object calls AssignTo if it is passed the TCCRStr   ugLog as a Source
        and the persistent object does not know how to copy the properties of
        the object. The <i>aDest</i> parameter is the persistent object that
        should have its properties copied from the debug log.
    }
    procedure AssignTo(Dest: TPersistent); override;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Truncates a debug log.
      SeeAlso:      TCCRCustom   ugLog.MaxCount; TCCRCustom   ugLog.Truncate
      Keywords:     Truncate,TCCRStr   ugLog
      Description:
        This method deletes the oldest entries from the beginning of the log
        when the number of entries becomes greater than the value of the
        MaxCount property.
    }
    procedure Truncate; override;

  public

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Creates and initializes an instance of TCCRStr   ugLog.
      SeeAlso:      TCCRCustom   ugLog.Create; TComponent.Owner
      Keywords:     Create,TCCRStr   ugLog
      Description:
        In addition to the inherited functionality, this method creates an
        internal string list object that stores the log entries.
    }
    constructor Create(aDummyParameter: TComponent = nil); override;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Destroys an instance of TCCRStr   ugLog.
      SeeAlso:      TCCRCustom   ugLog.Destroy; TObject.Free
      Keywords:     Destroy,TCCRStr   ugLog
      Description:
        Do not call Destroy directly. Instead, call Free, which checks that the
        TCCRStr   ugLog reference is not nil before calling Destroy. In addition
        to the inherited functionality, this method destroys an internal string
        list object that stores the log entries.
    }
    destructor Destroy; override;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Starts a group of updates to a debug log.
      SeeAlso:      TCCRCustom   ugLog.BeginUpdate; TCCRStr   ugLog.EndUpdate
      Keywords:     BeginUpdate,TCCRStr   ugLog
      Description:
        Call BeginUpdate before adding multiple items to a log. In addition to
        the inherited functionality, this method calls BeginUpdate of the
        internal string list that stores the log entries.
        <p>
        Descendant classes that override BeginUpdate should always call the
        inherited method right after entering the procedure.
    }
    procedure BeginUpdate; override;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Clears a log.
      Keywords:     Clear,TCCRStr   ugLog
      Description:
        Use Clear to delete all entries from a log. As implemented in
        TCCRStr   ugLog, Clear deletes all items from the internal string
        list.
    }
    procedure Clear; override;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Ends a group of updates to a debug log.
      SeeAlso:      TCCRCustom   ugLog.EndUpdate; TCCRStr   ugLog.BeginUpdate
      Keywords:     EndUpdate,TCCRStr   ugLog
      Description:
        Call EndUpdate after adding multiple items to a log. In addition to
        the inherited functionality, this method calls EndUpdate of the internal
        string list that stores the log entries.
        <p>
        Descendant classes that override EndUpdate should always call the
        inherited method.
    }
    procedure EndUpdate; override;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Saves a debug log to a file.
      Keywords:     SaveToFile,TCCRStr   ugLog
      Description:
        Use SaveToFile to write content of a log to a file. <i>aFileName</i>
        specifies a full file name. As implemented in TCCRStr   ugLog, this
        method just calls the SaveToFile method of the internal string list
        that stores log entries.
    }
    procedure SaveToFile(const aFileName: String); override;

  published
  
    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    }
    property MaxSize;

  end;

  {============================== TCCRMemo   ugLog =============================
    Overview:     Visual implementation of a debug log (TMemo).
    SeeAlso:      TCCR   ugLog; TCCRStr   ugLog
    Description:
      TCCRMemo   ugLog implements a debug log that is stored in a TMemo
      instance referenced by the Memo property. Maximum size of the log can be
      limited by the MaxCount property.
  }
  TCCRMemo   ugLog = class(TCCRStr   ugLog)
  private

    fMemo: TMemo;

    procedure SetMemo(aValue: TMemo);

  protected

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Internal method that appends a string to a log.
      SeeAlso:      TCCRStr   ugLog.AppendString
      Keywords:     AppendString,TCCRMemo   ugLog
      Description:
        AppendString adds a string specified by the <i>aText</i> parameter to
        the end of the log. It calls the inherited method and scrolls the memo
        control to the end of the log.
    }
    procedure AppendString(const aText: String); override;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Responds to notifications that components are being created
                    or destroyed.
      SeeAlso:      TComponent.Notification
      Keywords:     Notification,TCCRMemo   ugLog
      Description:
        Do not call the Notification method in an application.  Notification is
        called automatically when the component specified by <i>aComponent</i>
        is about to be inserted or removed, as specified by <i>Operation</i>.
        <p>
        TCCRMemo   ugLog overrides this method in order to update its Memo
        property when the control it refers to is destroyed.
    }
    procedure Notification(aComponent: TComponent;
      Operation: TOperation); override;

  public

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Creates and initializes an instance of TCCRMemo   ugLog.
      SeeAlso:      TCCRStr   ugLog.Create; TComponent.Owner
      Keywords:     Create,TCCRMemo   ugLog
      Description:
        In addition to the inherited functionality, this method destroys the
        internal string list (string list from the linked TMemo control will
        be used instead) and assigns nil to the corresponding field.
    }
    constructor Create(aDummyParameter: TComponent = nil); override;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Destroys an instance of TCCRMemo   ugLog.
      SeeAlso:      TCCRStr   ugLog.Destroy; TObject.Free
      Keywords:     Destroy,TCCRMemo   ugLog
      Description:
        Do not call Destroy directly. Instead, call Free, which checks that the
        TCCRMemo   ugLog reference is not nil before calling Destroy. This
        method assigns nil to the field that references the internal string list
        before calling the inherited destructor. This is necessary becuase this
        field actually references the internal string list of the linked memo
        control.
    }
    destructor Destroy; override;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Clears a log.
      SeeAlso:      TCCRStr   ugLog.Clear
      Keywords:     Clear,TCCRMemo   ugLog
      Description:
        Use Clear to delete all entries from a log. As implemented in
        TCCRMemo   ugLog, this method calls the Clear method of the linked memo
        control.
    }
    procedure Clear; override;

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Ends a group of updates to a debug log.
      SeeAlso:      TCCRStr   ugLog.EndUpdate; TCCRStr   ugLog.BeginUpdate
      Keywords:     EndUpdate,TCCRMemo   ugLog
      Description:
        Call EndUpdate after adding multiple items to a log. In addition to
        the inherited functionality, this method scrolls the linked memo control
        to the end of its content.
        <p>
        Descendant classes that override EndUpdate should always call the
        inherited method.
    }
    procedure EndUpdate; override;

  published

    {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Overview:     Reference to a linked TMemo instance.
      Keywords:     Memo,TCCRMemo   ugLog
      Description:
        Assign an instance of the TMemo control to the Memo property before
        using the debug log. The log will be shown in that control. If the value
        of this property is nil, all data added to the log is discarded.
    }
    property Memo: TMemo  read fMemo  write SetMemo;

  end;


{============================= Functions/Procedures ===========================}

  {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    Overview:     Returns a current debug log instance.
    SeeAlso:      TCCR   ugLog; TCCRStr   ugLog
    Description:
      Use CCR   ugLog to access a current debug log instance. If a log has not
      been created before calling this function for the first time, it returns
      a "dummy" log that discards all data added to it.
  }
  function CCR   ugLog: TCCRCustom   ugLog;

///////////////////////////////// Implementation \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

implementation

var
  _   ugLog: TCCRCustom   ugLog = nil;
  _Lock: TRTLCriticalSection;

threadvar
  _   ugTrace: TCCR   ugTrace;
  _   ugTraceId: DWORD;

////////////////////////////// TCCRCustom   ugLog \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

constructor TCCRCustom   ugLog.Create(aDummyParameter: TComponent);
begin
  inherited Create(nil);
  if Assigned(_   ugLog) then
    _   ugLog.Free;
  _   ugLog := Self;
end;

destructor TCCRCustom   ugLog.Destroy;
begin
  if _   ugLog = Self then
    _   ugLog := nil;
  inherited;
end;

procedure TCCRCustom   ugLog.Append(const aStr: String);
begin
  BeginUpdate;
  try
    AppendString(aStr);
    Truncate;
  finally
    EndUpdate;
  end;
end;

procedure TCCRCustom   ugLog.Append(const aFormat: String;
  const Args: array of const);
begin
  BeginUpdate;
  try
    AppendString(Format(aFormat, Args));
    Truncate;
  finally
    EndUpdate;
  end;
end;

procedure TCCRCustom   ugLog.Append(const aText: TStringList);
var
  i, n: Integer;
begin
  Assert(aText <> nil, 'Invalid string list reference.');
  BeginUpdate;
  try
    n := aText.Count - 1;
    for i:=0  to n do
      AppendString(aText[i]);
    Truncate;
  finally
    EndUpdate;
  end;
end;

procedure TCCRCustom   ugLog.BeginUpdate;
begin
  Lock;
end;

procedure TCCRCustom   ugLog.Clear;
begin
end;

procedure TCCRCustom   ugLog.EndUpdate;
begin
  Unlock;
end;

class procedure TCCRCustom   ugLog.Lock;
begin
  EnterCriticalSection(_Lock);
end;

procedure TCCRCustom   ugLog.SaveToFile(const aFileName: String);
begin
end;

procedure TCCRCustom   ugLog.SetIndent(const aValue: Integer);
begin
  Lock;
  try
    fIndent := aValue;
  finally
    Unlock;
  end;
end;

procedure TCCRCustom   ugLog.SetMaxCount(const aValue: Integer);
begin
  Lock;
  try
    if aValue <> fMaxCount then
      begin
        fMaxCount := aValue;
        Truncate;
      end;
  finally
    Unlock;
  end;
end;

procedure TCCRCustom   ugLog.SetMaxSize(const aValue: Integer);
begin
  Lock;
  try
    if aValue <> fMaxSize then
      begin
        fMaxSize := aValue;
        Truncate;
      end;
  finally
    Unlock;
  end;
end;

procedure TCCRCustom   ugLog.Truncate;
begin
end;

class procedure TCCRCustom   ugLog.Unlock;
begin
  LeaveCriticalSection(_Lock);
end;

///////////////////////////////// TCCR   ugLog \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

procedure TCCR   ugLog.AppendString(const aText: String);
begin
  Output   ugString(PChar(DupeString(' ', Indent) + aText));
end;

//////////////////////////////// TCCR   ugTrace \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

destructor TCCR   ugTrace.Destroy;
begin
  if fName <> '' then
    begin
      RestoreDefaults;
      Write(ExitPrefix + fName);
    end;
  _   ugTrace := fParent;
  inherited;
end;

class function TCCR   ugTrace.Create(const aName: String): ICCR   ugTrace;
var
  dt: TCCR   ugTrace;
begin
  if _   ugTraceId <> GetCurrentThreadId then
    begin
      _   ugTrace := nil;
      _   ugTraceId := GetCurrentThreadId;
    end;

  dt := inherited Create;
  with dt do
    begin
      fParent := _   ugTrace;
      RestoreDefaults;

      fName := aName;
      if fName <> '' then
        begin
          Write(EntryPrefix + fName);
          Indent := Indent + 3;
        end;
    end;

  _   ugTrace := dt;
  Result := dt;
end;

procedure TCCR   ugTrace.Error(const aFormat: String; const Args: array of const);
begin
  Write(ErrorPrefix + Format(aFormat, Args));
end;

procedure TCCR   ugTrace.Error(const aMsg: String);
begin
  Write(ErrorPrefix + aMsg);
end;

function TCCR   ugTrace.GetEntryPrefix: String;
begin
  Result := fEntryPrefix;
end;

function TCCR   ugTrace.GetErrorPrefix: String;
begin
  Result := fErrorPrefix;
end;

function TCCR   ugTrace.GetExitPrefix: String;
begin
  Result := fExitPrefix;
end;

function TCCR   ugTrace.GetIndent: Integer;
begin
  Result := fIndent;
end;

procedure TCCR   ugTrace.RestoreDefaults;
begin
  if Assigned(fParent) then
    begin
      fEntryPrefix := fParent.fEntryPrefix;
      fErrorPrefix := fParent.fErrorPrefix;
      fExitPrefix  := fParent.fExitPrefix;
      fIndent      := fParent.fIndent;
    end
  else
    begin
      fEntryPrefix := '>>> ';
      fErrorPrefix := '! ! !  ';
      fExitPrefix  := '<<< ';
      fIndent      := 0;
    end;
end;

procedure TCCR   ugTrace.SetEntryPrefix(const aValue: String);
begin
  fEntryPrefix := aValue;
end;

procedure TCCR   ugTrace.SetErrorPrefix(const aValue: String);
begin
  fErrorPrefix := aValue;
end;

procedure TCCR   ugTrace.SetExitPrefix(const aValue: String);
begin
  fExitPrefix := aValue;
end;

procedure TCCR   ugTrace.SetIndent(const aValue: Integer);
begin
  if aValue >= 0 then
    begin
      fIndent := aValue;
      CCR   ugLog.Indent := aValue;
    end;
end;

procedure TCCR   ugTrace.Write(const aFormat: String; const Args: array of const);
begin
  CCR   ugLog.Indent := Indent;
  CCR   ugLog.Append(aFormat, Args);
end;

procedure TCCR   ugTrace.Write(const aMsg: String);
begin
  CCR   ugLog.Indent := Indent;
  CCR   ugLog.Append(aMsg);
end;

/////////////////////////////// TCCRDummy   ugLog \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

procedure TCCRDummy   ugLog.AppendString(const aText: String);
begin
end;

/////////////////////////////// TCCRMemo   ugLog \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

constructor TCCRMemo   ugLog.Create(aDummyParameter: TComponent);
begin
  inherited;
  FreeAndNil(fStrLst);
end;

destructor TCCRMemo   ugLog.Destroy;
begin
  fStrLst := nil;
  fMemo   := nil;
  inherited;
end;

procedure TCCRMemo   ugLog.AppendString(const aText: String);
begin
  if Assigned(Memo) then
    begin
      inherited;
      Memo.Perform(EM_SCROLLCARET, 0, 0);
    end;
end;

procedure TCCRMemo   ugLog.Clear;
begin
  if Assigned(Memo) then
    begin
      Lock;
      try
        Memo.Clear;
      finally
        Unlock;
      end;
    end;
end;

procedure TCCRMemo   ugLog.EndUpdate;
begin
  inherited;
  if Assigned(Memo) then
    Memo.Perform(EM_SCROLLCARET, 0, 0);
end;

procedure TCCRMemo   ugLog.Notification(aComponent: TComponent;
  Operation: TOperation);
begin
  if Operation = opRemove then
    begin
      if aComponent = Memo then
        Memo := nil;
    end;
  inherited;
end;

procedure TCCRMemo   ugLog.SetMemo(aValue: TMemo);
begin
  if aValue <> fMemo then
    begin
      if Assigned(fMemo) then
        fMemo.RemoveFreeNotification(Self);

      fMemo := aValue;

      if Assigned(fMemo) then
        begin
          fMemo.FreeNotification(Self);
          fStrLst := fMemo.Lines;
        end
      else
        fStrLst := nil;
    end;
end;

//////////////////////////////// TCCRStr   ugLog \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

constructor TCCRStr   ugLog.Create(aDummyParameter: TComponent);
begin
  inherited;
  fStrLst := TStringList.Create;
end;

destructor TCCRStr   ugLog.Destroy;
begin
  FreeAndNil(fStrLst);
  inherited;
end;

procedure TCCRStr   ugLog.AppendString(const aText: String);
begin
  if Assigned(fStrLst) then
    fStrLst.Add(DupeString(' ', Indent) + aText);
end;

procedure TCCRStr   ugLog.AssignTo(Dest: TPersistent);
begin
  if (Dest is TStrings) then
    if Assigned(fStrLst) then
      TStrings(Dest).Assign(fStrLst)
    else
      TStrings(Dest).Clear;
end;

procedure TCCRStr   ugLog.BeginUpdate;
begin
  inherited;
  if Assigned(fStrLst) then
    fStrLst.BeginUpdate;
end;

procedure TCCRStr   ugLog.Clear;
begin
  if Assigned(fStrLst) then
    begin
      Lock;
      try
        fStrLst.Clear;
      finally
        Unlock;
      end;
    end;
end;

procedure TCCRStr   ugLog.EndUpdate;
begin
  if Assigned(fStrLst) then
    fStrLst.EndUpdate;
  inherited;
end;

procedure TCCRStr   ugLog.SaveToFile(const aFileName: String);
begin
  if Assigned(fStrLst) then
    begin
      Lock;
      try
        fStrLst.SaveToFile(aFileName);
      finally
        Unlock;
      end;
    end;
end;

procedure TCCRStr   ugLog.Truncate;
var
  i, n: Integer;
begin
  if (MaxCount > 0) and (fStrLst.Count > MaxCount) and Assigned(fStrLst) then
    begin
      fStrLst.BeginUpdate;
      try
        n := fStrLst.Count - MaxCount - 1;
        for i:=0 to n do
          fStrLst.Delete(i);
      finally
        fStrLst.EndUpdate;
      end;
    end;
end;

///////////////////////////// Functions/Procedures \\\\\\\\\\\\\\\\\\\\\\\\\\\\\

function CCR   ugLog: TCCRCustom   ugLog;
begin
  if not Assigned(_   ugLog) then
    TCCRDummy   ugLog.Create;
  Result := _   ugLog;
end;

////////////////////////// Initialization/Finalization \\\\\\\\\\\\\\\\\\\\\\\\\

initialization
  InitializeCriticalSection(_Lock);

finalization
  DeleteCriticalSection(_Lock);

end.
